﻿/*
 * Copyright (c) 2013-present, The Eye Tribe. 
 * All rights reserved.
 *
 * This source code is licensed under the BSD-style license found in the LICENSE file in the root directory of this source tree. 
 *
 */

using UnityEngine;
using System.Collections;
using TETCSharpClient;
using TETCSharpClient.Data;
using Assets.Scripts;

/// <summary>
/// Component attached to 'Main Camera' of '/Scenes/std_scene.unity'.
/// This script handles the navigation of the 'Main Camera' according to 
/// the GazeData stream recieved by the EyeTribe Server.
/// </summary>
public class GazeCamera : MonoBehaviour, IGazeListener
{
    private Camera cam;

    private double eyesDistance;
    private double baseDist;
    private double depthMod;

    private Component gazeIndicator;

    private Collider currentHit;

    private GazeDataValidator gazeUtils;
	//
	private int currentScore;
	private int scoreMax;
	//
	private double angle;
	private GameObject plateau;
	private GameObject targets;
	private float _angleWindow;


    void Start()
    {
        //Stay in landscape
        Screen.autorotateToPortrait = false;
		plateau = GameObject.Find ("Plateau");
		targets = GameObject.Find ("targets");
        cam = GetComponent<Camera>();
        baseDist = cam.transform.position.z;
        gazeIndicator = cam.transform.GetChild(0);
		_angleWindow = 5.0F;

        //initialising GazeData stabilizer
        gazeUtils = new GazeDataValidator(30);

        //register for gaze updates
        GazeManager.Instance.AddGazeListener(this);
		//
		currentScore=0;
		scoreMax = 0;
    }

    public void OnGazeUpdate(GazeData gazeData)
    {
        //Add frame to GazeData cache handler
        gazeUtils.Update(gazeData);
    }

    void Update()
    {
        Point2D userPos = gazeUtils.GetLastValidUserPosition();

        if (null != userPos)
        {
            //mapping cam panning to 3:2 aspect ratio
            double tx = (userPos.X * 5) - 2.5f;
            double ty = (userPos.Y * 3) - 1.5f;

            //position camera X-Y plane and adjust distance
            eyesDistance = gazeUtils.GetLastValidUserDistance();
            depthMod = 2 * eyesDistance;

            Vector3 newPos = new Vector3(
                (float)tx,
                (float)ty,
                (float)(baseDist + depthMod));
            cam.transform.position = newPos;

            //camera 'look at' origo
            cam.transform.LookAt(Vector3.zero);

            //tilt cam according to eye angle
          
			angle= gazeUtils.GetLastValidEyesAngle();
           // cam.transform.eulerAngles = new Vector3(cam.transform.eulerAngles.x, cam.transform.eulerAngles.y, cam.transform.eulerAngles.z + (float)angle);
			cam.transform.eulerAngles = new Vector3(0, 0, cam.transform.eulerAngles.z + (float)angle);
			if (angle > _angleWindow || angle < -_angleWindow) {
				targets.SetActive (false);
			} else if ((!targets.activeSelf && angle < _angleWindow) || (!targets.activeSelf && angle > -_angleWindow)) {
				targets.SetActive (true);
			}
				
        }

        Point2D gazeCoords = gazeUtils.GetLastValidSmoothedGazeCoordinates();

        if (null != gazeCoords)
        {
            //map gaze indicator
            Point2D gp = UnityGazeUtils.getGazeCoordsToUnityWindowCoords(gazeCoords);

            Vector3 screenPoint = new Vector3((float)gp.X, (float)gp.Y, cam.nearClipPlane + .1f);

            Vector3 planeCoord = cam.ScreenToWorldPoint(screenPoint);
            gazeIndicator.transform.position = planeCoord;

            //handle collision detection
            checkGazeCollision(screenPoint);
        }

        //handle keypress
        if (Input.GetKey(KeyCode.Escape))
        {
            Application.Quit();
        }
        else if (Input.GetKey(KeyCode.Space))
        {
            Application.LoadLevel(0);
        }
		else if (Input.GetKey(KeyCode.UpArrow))
		{
			float z = plateau.transform.position.z;
			plateau.transform.position=new Vector3(0, 0, z+.1f);
		}
		else if (Input.GetKey(KeyCode.DownArrow))
		{
			float z = plateau.transform.position.z;
			plateau.transform.position=new Vector3(0, 0, z-.1f);
		}
		//currentScore++;
		//scoreMax++;
		if (currentScore > scoreMax) {
			scoreMax = currentScore;
		}
    }

    private void checkGazeCollision(Vector3 screenPoint)
    {
        Ray collisionRay = cam.ScreenPointToRay(screenPoint);
        RaycastHit hit;

        if (Physics.Raycast(collisionRay, out hit))
        {
			
            if (null != hit.collider && currentHit != hit.collider)
            {
                //switch colors of cubes according to collision state

                if (null != currentHit)
                    currentHit.GetComponent<Renderer>().material.color = Color.white;
                currentHit = hit.collider;
				if (currentHit.name == "target") {
					currentHit.GetComponent<Renderer>().material.color = Color.green;

				} else {
					currentHit.GetComponent<Renderer>().material.color = Color.red;
					currentScore = 0;
				}

            }
			if (currentHit.name == "target") {
				currentScore++;
			}
        }
    }

    void OnGUI()
    {
        int padding = 10;
        int btnWidth = 160;
        int btnHeight = 40;
        int y = padding;

        if (GUI.Button(new Rect(padding, y, btnWidth, btnHeight), "Quitter"))
        {
            Application.Quit();
        }

        y += padding + btnHeight;

        if (GUI.Button(new Rect(padding, y, btnWidth, btnHeight), "Commencer"))
        {
            Application.LoadLevel(0);
        }

		//
		y += (padding + btnHeight)*2;
		GUI.Box(new Rect(padding, y, btnWidth, btnHeight), "Score : "+currentScore);
		y += (padding + btnHeight);
		GUI.Box(new Rect(padding, y, btnWidth, btnHeight), "Score Max : "+scoreMax);
		y += (padding + btnHeight);
		if (GUI.Button(new Rect(padding, y, btnWidth, btnHeight), "Reset Score Max"))
		{
			scoreMax = 0;
		}
		//
		y += (padding + btnHeight)*4;
		if (GUI.Button(new Rect(padding, y, btnWidth, btnHeight), "Avancer le plateau"))
		{
			float z = plateau.transform.position.z;
			plateau.transform.position=new Vector3(0, 0, (z-.1f));
		}
		y += (padding + btnHeight);
		if (GUI.Button(new Rect(padding, y, btnWidth, btnHeight), "Reculer le plateau"))
		{
			float z = plateau.transform.position.z;
			plateau.transform.position=new Vector3(0, 0, z+.1f);
		}
		y += (padding + btnHeight)*2;
		GUI.Box(new Rect(padding, y, btnWidth, btnHeight), "Angle : "+angle.ToString("n1"));
		y += (padding + btnHeight);
		GUI.Box(new Rect(padding, y, btnWidth, btnHeight), "Angle max : "+_angleWindow.ToString("n1"));
		y += (padding + btnHeight);
		GUI.backgroundColor=Color.white;
		_angleWindow=GUI.HorizontalSlider(new Rect(padding, y, btnWidth, btnHeight),_angleWindow, 0.0F, 20.0F);
    }

    void OnApplicationQuit()
    {
        GazeManager.Instance.RemoveGazeListener(this);
    }
}
